@@ -3,4 +3,5 @@ from ncem.interpretation.interpreter import (InterpreterCVAEncem,
                                              InterpreterEDncem,
                                              InterpreterGraph,
                                              InterpreterInteraction,
-                                             InterpreterLinear)
+                                             InterpreterLinear,
+                                             InterpreterDeconvolution)
