@@ -1,3 +1,4 @@
+"""Importing custom layers for different model classes."""
 from ncem.models.layers.gnn_layers import GCNLayer, MaxLayer
 from ncem.models.layers.layer_stacks_lvm import (CondDecoder, CondEncoder,
                                                  Decoder, Encoder,
@@ -8,6 +9,7 @@ from ncem.models.layers.output_layers import (GaussianConstDispOutput,
                                               LinearOutput,
                                               NegBinConstDispOutput,
                                               NegBinOutput,
-                                              NegBinSharedDispOutput)
-from ncem.models.layers.preproc_input import (DenseInteractions, NodeDegrees,
-                                              PreprocInput)
+                                              NegBinSharedDispOutput,
+                                              get_out)
+from ncem.models.layers.preproc_input import DenseInteractions, PreprocInput
+from ncem.models.layers.single_gnn_layers import SingleLrGatLayer, SingleGatLayer
