@@ -2,4 +2,4 @@
 from ncem.train.summaries import GridSearchContainer
 from ncem.train.train_model import (TrainModelCVAE, TrainModelCVAEncem,
                                     TrainModelED, TrainModelEDncem, TrainModelEdSingleNcem,
-                                    TrainModelInteractions, TrainModelLinear)
+                                    TrainModelInteractions, TrainModelLinear, TrainModelLinearDeconvolution)
